
# Load Python libraries that will be used in this notebook
from azure.common.client_factory import get_client_from_cli_profile
from azure.common.credentials import get_azure_cli_credentials
from azure.loganalytics.models import QueryBody
from azure.mgmt.loganalytics import LogAnalyticsManagementClient
from azure.loganalytics import LogAnalyticsDataClient

# Processing Helpers
import pandas as pd
import numpy as np
import ipywidgets
from IPython.display import display, HTML, Markdown
import json
import networkx as nx
import csv
import requests
import jsonpickle

from io import BytesIO
import msal
import logging

#Bokeh Visualization Helpers
import bokeh.io
from bokeh.io import output_notebook, show, save
from bokeh.models import Range1d, Circle, ColumnDataSource, MultiLine
from bokeh.plotting import figure, from_networkx
from bokeh.palettes import RdYlGn11
from bokeh.transform import linear_cmap
from bokeh.resources import INLINE
bokeh.io.output_notebook(INLINE)
import scipy
import tkinter as tk

# Globals
token = None
graphApiVersion = "beta"
uri = "https://graph.microsoft.com/{v}/{r}"
headers = None

# Functions
def authenticate(tenant: str, id: str, secret: str):
    '''Authenticate against the Graph API, return a token'''

    global token
    global headers
    authority = "https://login.microsoftonline.com/{0}".format(tenant)
    appID = id
    appSecret = secret
    scope = ["https://graph.microsoft.com/.default"]

    app = msal.ConfidentialClientApplication(
        appID, authority=authority, client_credential = appSecret)
    token = app.acquire_token_silent(scope, account=None)
    if not token:
        token = app.acquire_token_for_client(scopes=scope)
    headers = {'Authorization': 'Bearer ' + token['access_token']}
    return

def users():
    '''Fetch Users data from the Graph API'''
    return query(graphApiVersion, "users?")

def usersMemberships():
    '''Fetch User Memberships data from the Graph API'''
    return query(graphApiVersion, "users?$expand=memberOf")

def usersManagers():
    '''Fetch User Manager data from the Graph API'''
    return query(graphApiVersion, "users?$expand=manager")

def groups():
    '''Fetch Groups data from the Graph API'''
    return query(graphApiVersion, "groups")

def groupsMemberships():
    '''Fetch Group Memberships data from the Graph API'''
    return query(graphApiVersion, "groups?$expand=members")

def groupsOwners():
    '''Fetch Group Owners data from the Graph API'''
    return query(graphApiVersion, "groups?$expand=owners")

def roles():
    '''Fetch Roles data from the Graph API'''
    return query(graphApiVersion, "directoryRoles")

def rolesMembers():
    '''Fetch Role Members data from the Graph API'''
    return query(graphApiVersion, "directoryRoles?$expand=members")

def applications():
    '''Fetch Applications data from the Graph API'''
    return query(graphApiVersion, "applications")

def applicationsOwners():
    '''Fetch Application Owners data from the Graph API'''
    return query(graphApiVersion, "applications?$expand=owners")

def servicePrincipals():
    '''Fetch Servicfe Principals data from the Graph API'''
    return query(graphApiVersion, "servicePrincipals")

def servicePrincipalsOwners():
    '''Fetch Service Pricnipal Owners data from the Graph API'''
    return query(graphApiVersion, "servicePrincipals?$expand=owners")

def query(v, r):
    '''Fetch data from the Graph API, given an endpoint'''

    dest = uri.format(v=v, r=r)
    result = requests.get(dest, headers=headers).json()

    test = getNextLinkQuery(result)

    while test is not None:
        result["value"].extend(test["value"])
        test = getNextLinkQuery(test)
        
    return result

def getNextLinkQuery(result):
    '''Fetch the next page of data from the Graph API'''

    if "@odata.nextLink" in result:
        newResult = requests.get(result["@odata.nextLink"], headers=headers).json()
        return newResult
    return

# *********************

def PluckDataFromQueryResults(userMemberDict, userManagerDict, groupOwnerDict, groupMemberDict, roleMemberDict, appOwnerDict, spDict, filewriter): 
    '''Process data from Permission Query'''

    filewriter.writerow(['Source', 'SourceId', 'SourceType', 'Target', 'TargetId', 'TargetType', 'SourceWeight', 'TargetWeight', 'Relationship', 'Reason', 'Risk', 'UserMail', 'UserManagerMail'])

    # Set up some values
    role_member_relationship = 'Role Member'
    group_member_relationship = 'Group Member'
    group_owner_relationship = 'Group Owner'
    app_owner_relationship = 'Application Owner'
    service_principal_owner_relationship = 'Service Principal Owner'
    manager_relationship = 'Manager'
    unknown_relationship = 'Unknown'

    very_low_risk = 'VeryLow'
    low_risk = 'Low'
    medium_risk = 'Medium'
    high_risk = 'High'
    very_high_risk = 'VeryHigh'
    

    weight_user = '1'
    weight_disabled_user = '3'
    weight_group = '4'
    weight_role = '5'
    weight_app = '9'
    weight_unknown = '10'

    user_type = "User"
    disabled_user_type = "Disabled User"
    group_type = "Group"
    role_type = "Role"
    app_type = "Application"
    service_principal_type = "Service Principal"
    unknown_type = "Unknown"
    
    userList = []
    disabledUserList = []
    unknownObjectList = []
    directGroupList = []
    directRoleList = []
    directApplicationList = []
    directServicePrincipalList = []
    directUnknownList = []
    indirectGroupList = []
    indirectRoleList = []
    indirectUnknownList = []
    edges = []

    if "value" in userMemberDict:
        for user in userMemberDict["value"]:
            processingHelper(user["displayName"])
            userMail = user['mail']
            userManagerMail = ''
            risk = low_risk
            weight = MapRiskToValue(risk)
            usertype = user_type

            if user["accountEnabled"] == True:
                userList.append(user["displayName"])
            else:
                disabledUserList.append(user["displayName"])
                usertype = disabled_user_type
                weight = weight_disabled_user
                risk = medium_risk
                weight = MapRiskToValue(risk)

            if "memberOf" in user:
                for userMembership in user["memberOf"]:
                    if userMembership["@odata.type"] == "#microsoft.graph.directoryRole":
                        directRoleList.append(userMembership["displayName"])

                        if("Global Administrator" in userMembership["displayName"]):
                            risk = high_risk
                            weight = MapRiskToValue(risk)
                        elif("Administrator" in userMembership["displayName"]):
                            risk = medium_risk
                            weight = MapRiskToValue(risk)

                        edges.append([user["displayName"], user["id"], usertype, userMembership["displayName"],userMembership["id"], role_type,weight, weight, role_member_relationship, role_member_relationship, risk, userMail, userManagerMail])
                    elif userMembership["@odata.type"] == "#microsoft.graph.group":
                        directGroupList.append(userMembership["displayName"])
                        edges.append([user["displayName"], user["id"],usertype,userMembership["displayName"],userMembership["id"],group_type,weight, weight, group_member_relationship, group_member_relationship, risk, userMail, userManagerMail])
                    else:
                        unknownObjectList.append(userMembership["id"])
                        edges.append([user["displayName"], user["id"],usertype,userMembership["displayName"],userMembership["id"],unknown_type, weight, weight, unknown_relationship, unknown_relationship, medium_risk, userMail, userManagerMail])

    if "value" in userManagerDict:
        for user in userManagerDict["value"]:
            processingHelper(user["displayName"])
            userMail = user['mail']
            userManagerMail = ''
            risk = very_low_risk
            weight = MapRiskToValue(risk)
            usertype = user_type
            
            if user["accountEnabled"] == True:
                userList.append(user["displayName"])
                weight = weight_user
            else:
                disabledUserList.append(user["displayName"])
                usertype = disabled_user_type
                weight = weight_disabled_user
                risk = low_risk
                weight = MapRiskToValue(risk)

            if "manager" in user:
                userManager = user["manager"]
                if "mail" in userManager:
                    userManagerMail = userManager["mail"]
                edges.append([user["displayName"], user["id"],usertype,userManager["displayName"],userManager["id"],user_type,weight, weight, manager_relationship, manager_relationship, risk, userMail, userManagerMail])

    if "value" in groupOwnerDict:
        for group in groupOwnerDict["value"]:
            processingHelper(group["displayName"])
            groupMail = group['mail']
            groupManagerMail = group['mail']
            directGroupList.append(group["displayName"])
            risk = medium_risk
            weight = MapRiskToValue(risk)

            for owner in group["owners"]:
                for user in userMemberDict["value"]:
                    if owner["id"] == user["id"]:
                        usertype = user_type
                        if user["accountEnabled"] == True:
                            userList.append(user["displayName"])
                            weight = MapRiskToValue(risk)
                        else:
                            disabledUserList.append(user["displayName"])
                            usertype = disabled_user_type
                            weight = weight_disabled_user
                            risk = high_risk
                            weight = MapRiskToValue(risk)
                        edges.append([group["displayName"], group["id"],group_type,user["displayName"],user["id"],usertype,weight, weight, group_owner_relationship, group_owner_relationship, risk, groupMail, groupManagerMail])

    if "value" in groupMemberDict:
        for group in groupMemberDict["value"]:
            processingHelper(group["displayName"])
            groupMail = group['mail']
            groupManagerMail = group['mail']
            directGroupList.append(group["displayName"])
            risk = low_risk
            weight = MapRiskToValue(risk)

            for member in group["members"]:
                for user in userMemberDict["value"]:
                    if member["id"] == user["id"]:
                        usertype = user_type
                        if user["accountEnabled"] == True:
                            userList.append(user["displayName"])
                            weight = MapRiskToValue(risk)
                        else:
                            disabledUserList.append(user["displayName"])
                            usertype = disabled_user_type
                            risk = medium_risk
                            weight = MapRiskToValue(risk)
                        edges.append([group["displayName"], group["id"],group_type,user["displayName"],user["id"],usertype,weight, weight, group_member_relationship, group_member_relationship, risk, groupMail, groupManagerMail])
                for innerGroup in groupMemberDict["value"]:
                    if member["id"] == innerGroup["id"]:
                        directGroupList.append(innerGroup["displayName"])
                        risk = medium_risk
                        weight = MapRiskToValue(risk)
                        edges.append([group["displayName"], group["id"],group_type,innerGroup["displayName"],innerGroup["id"],group_type,weight, weight, group_member_relationship, group_member_relationship, risk, groupMail, groupManagerMail])

    if "value" in roleMemberDict:
        for role in roleMemberDict["value"]:
            processingHelper(role["displayName"])
            roleMail = ''
            roleManagerMail = ''
            directRoleList.append(role["displayName"])

            if("Global Administrator" in role["displayName"]):
                risk = high_risk
            elif("Administrator" in role["displayName"]):
                risk = medium_risk
            else:
                risk = low_risk

            weight = MapRiskToValue(risk)
                
            for member in role["members"]:
                for user in userMemberDict["value"]:
                    if member["id"] == user["id"]:
                        usertype = user_type
                        if user["accountEnabled"] == True:
                            userList.append(user["displayName"])
                        else:
                            disabledUserList.append(user["displayName"])
                            usertype = disabled_user_type
                            weight = weight_disabled_user

                            if("Global Administrator" in role["displayName"]):
                                risk = very_high_risk
                            elif("Administrator" in role["displayName"]):
                                risk = high_risk
                            else:
                                risk = medium_risk
                            
                            weight = MapRiskToValue(risk)

                        edges.append([role["displayName"], role["id"],role_type,user["displayName"],user["id"],usertype,weight_role, weight, role_member_relationship, role_member_relationship, risk, roleMail, roleManagerMail])
                for group in groupMemberDict["value"]:
                    if member["id"] == group["id"]:
                        directGroupList.append(group["displayName"])
                        weight = weight_group
                        risk = medium_risk
                        edges.append([role["displayName"], role["id"],role_type,group["displayName"],group["id"],group_type,weight_role, weight, role_member_relationship, role_member_relationship, risk, roleMail, roleManagerMail])

    if "value" in appOwnerDict:
        for application in appOwnerDict["value"]:
            processingHelper(application["displayName"])
            appMail = ''
            appManagerMail = ''
            risk = low_risk
            directApplicationList.append(application["displayName"])
            weight = MapRiskToValue(risk)

            for owner in application["owners"]:
                for user in userMemberDict["value"]:
                    if owner["id"] == user["id"]:
                        usertype = user_type
                        if user["accountEnabled"] == True:
                            userList.append(user["displayName"])
                        else:
                            disabledUserList.append(user["displayName"])
                            usertype = disabled_user_type
                            risk = high_risk
                        
                        weight = MapRiskToValue(risk)
                        edges.append([application["displayName"], application["id"],app_type,user["displayName"],user["id"],usertype,weight, weight, app_owner_relationship, app_owner_relationship, risk, appMail, appManagerMail])


    if "value" in spDict:
        for servicePrincipal in spDict["value"]:
            processingHelper(servicePrincipal["displayName"])
            spMail = ''
            spManagerMail = ''
            risk = medium_risk
            directServicePrincipalList.append(servicePrincipal["displayName"])
            weight = MapRiskToValue(risk)

            for owner in servicePrincipal["owners"]:
                for user in userMemberDict["value"]:
                    if owner["id"] == user["id"]:
                        usertype = user_type
                        if user["accountEnabled"] == True:
                            userList.append(user["displayName"])
                        else:
                            disabledUserList.append(user["displayName"])
                            usertype = disabled_user_type
                            risk = high_risk
                        
                        weight = MapRiskToValue(risk)
                        edges.append([servicePrincipal["displayName"], servicePrincipal["id"],service_principal_type, user["displayName"],user["id"],usertype,weight, weight, service_principal_owner_relationship, service_principal_owner_relationship, risk, spMail, spManagerMail])

    defaultFilter = '(no filter applied)'
    dropdown = [defaultFilter]
    for user in userList:
        dropdown.append(user)
    for disabledUser in disabledUserList:
        dropdown.append(disabledUser)
    for group in directGroupList:
        dropdown.append(group)
    for role in directRoleList:
        dropdown.append(role)
    for servicePrincipal in directServicePrincipalList:
        dropdown.append(servicePrincipal)
    for application in directApplicationList:
        dropdown.append(application)
    for unknown in directUnknownList:
        dropdown.append(unknown)
    for edge in edges:
        filewriter.writerow(edge)

    return

def filterHelper(name_dropdown, file_df, defaultFilter):
    '''Based on given edge data file and dropdown selection, filter the edge data'''

    if(name_dropdown.value != defaultFilter):
        file_df_filtered = file_df[(file_df.Target == name_dropdown.value)]
        file_df_filtered = file_df_filtered.append(file_df[(file_df.Source == name_dropdown.value)])
        return file_df_filtered
    else:
        return file_df
   

def processingHelper(item: str):
    '''Status dialog to command line'''
    print("Processing " + item + "...")

def glyph_map_helper(item:str):
    '''Mapper for glyphs to type in Graph Renderer'''
    dict = {'User': "circle", 'Group': "diamond", 'Application': "triangle", 'Service Principal': "star", 'Disabled User': "plus", 'PIM Role': "hex", 'Role': "square", 'Default': "circle_x"}

    return dict[item] if item in dict else dict["Default"]

def RenderGraphData(file_df):
    '''Given a Pandas data frame from a CSV file, parse the data out, process it for rendering with a Networkx Graph.
    Each line in Data frame is expected to be in form [Source,Target,SourceWeight,TargetWeight,Relationship,Risk]'''

    # Set up the Graph 
    G = nx.Graph
    keys = {}
    values = {}
    modularity_class = {}
    modularity_color = {}
    modularity_glyph = {}
    relationship = {}
    risk = {}
    target_id = {}
    target_type = {}

    # Process the passed in data frame and make nodes/edges data for graph
    # Use a data frame because networkx expects a data frame
    G = nx.from_pandas_edgelist(file_df, 'Source','Target', True)
    for index, row in file_df.iterrows():
        # SSame node might appear multiple times, find and use only highest value
        if(row['Source'] in keys):
            keys[row['Source']] = max(int(row['SourceWeight']), int(keys[row['Source']]))
        else:
            keys[row['Source']] = int(row['SourceWeight'])

        if(row['Target'] in keys):
            keys[row['Target']] = max(int(row['TargetWeight']), int(keys[row['Target']]))
        else:
            keys[row['Target']] = int(row['TargetWeight'])

        modularity_class[row['Source']] = keys[row['Source']]
        modularity_color[row['Source']] = RdYlGn11[keys[row['Target']]]
        modularity_class[row['Target']] = keys[row['Source']]
        modularity_color[row['Target']] = RdYlGn11[keys[row['Target']]]
        modularity_glyph[row['Source']] = glyph_map_helper(row['SourceType'])
        modularity_glyph[row['Target']] = glyph_map_helper(row['TargetType'])
        relationship[row['Target']] = row['Relationship']
        target_id[row['Source']] = row['SourceId']
        target_id[row['Target']] = row['TargetId']
        target_type[row['Source']] = row['SourceType']
        target_type[row['Target']] = row['TargetType']

        if(row['Source'] in risk.keys()):
            risk[row['Source']] = RiskComparer(risk[row['Source']], row['Risk'])
        else:
            risk[row['Source']] = row['Risk']

        if(row['Target'] in risk.keys()):
            risk[row['Target']] = RiskComparer(risk[row['Target']], row['Risk'])
        else:
            risk[row['Target']] = row['Risk']

    degrees = dict(nx.degree(G))
    nx.set_node_attributes(G, degrees, 'degree')

    # Set the node size after we have processed the data
    number_to_adjust_by = 10
    adjusted_node_size = dict([(node, degree+number_to_adjust_by) for node, degree in nx.degree(G)])
    nx.set_node_attributes(G, adjusted_node_size, 'adjusted_node_size')

    # Add modularity class, color, and relationship as attributes from the network above
    nx.set_node_attributes(G, modularity_class, 'modularity_class')
    nx.set_node_attributes(G, modularity_color, 'modularity_color')
    nx.set_node_attributes(G, relationship, 'relationship')
    nx.set_node_attributes(G, risk, 'risk')
    nx.set_node_attributes(G, target_id, 'target_id')
    nx.set_node_attributes(G, target_type, 'target_type')
    nx.set_node_attributes(G, modularity_glyph, 'marker_type')



    #Choose attributes from G network to size and color by — setting manual size (e.g. 10) or color (e.g. 'skyblue') also allowed
    size_by_this_attribute = 'adjusted_node_size'
    color_by_this_attribute = 'modularity_color'
    marker_by_this_attribute = 'marker_type'

    #Choose a title!
    title = 'Needle in the Haystack'

    #Establish which categories will appear when hovering over each node
    HOVER_TOOLTIPS = [
        ("Azure Name", "@index"),
        ("Azure ID", "@target_id"),
        ("Azure Object Type", "@target_type"),
        ("Number of Connections", "@degree"),
        ("Risk", "@risk"),
    ]

    #Create a plot — set dimensions, toolbar, and title
    plot = figure(tooltips = HOVER_TOOLTIPS,
                tools="pan,wheel_zoom,save,reset, tap", active_scroll='wheel_zoom',
                x_range=Range1d(-10.1, 10.1), y_range=Range1d(-10.1, 10.1), title=title)

    # Create a network graph object
    # https://networkx.github.io/documentation/networkx-1.9/reference/generated/networkx.drawing.layout.spring_layout.html
    network_graph = from_networkx(G, nx.spring_layout, scale=10, center=(0, 0))

    #Set node sizes and colors according to node degree (color as category from attribute)
    network_graph.node_renderer.glyph = Scatter(size=size_by_this_attribute, fill_color=color_by_this_attribute, marker=marker_by_this_attribute)
    
    # Set edge opacity and width
    network_graph.edge_renderer.glyph = MultiLine(line_alpha=0.5, line_width=1)

    # Render the graph
    plot.renderers.append(network_graph)
    
    return plot
    
def SenservaPermissionQuery(tableName):
    '''Query for finding relationships from the Senserva Scanner results. 
    If you have set up multiple workspaces, provide a Sentinel alias to query both functions'''

    where_clause = "{0} | where TimeGenerated > ago(7d)| where (ControlName_s == 'ServicePrincipalPermissionGrantTenant' or ControlName_s == 'ApplicationPermissionGrantTenant' or ControlName_s == 'ServicePrincipalMembership' or ControlName_s == 'UserMembers' or ControlName_s =='UserOwners' or ControlName_s =='GroupMembers' or ControlName_s == 'UserManagers') | extend values =tostring(parse_json(Value_s)) | extend JSON = todynamic(values) | order by TimeGenerated desc"
    return where_clause.format(tableName)

def ParseQueryResultHelper(nameValuePair, foundList, defaultList, edgesList, primaryKey, backupKey, sourceName, weightSource, weightFound, weightNotFound, relationship, risk, userMail,userManagerMail):
    '''Helper to parse through query results'''

    if(primaryKey in nameValuePair.keys()):
        foundList.append(nameValuePair[primaryKey])
        edgesList.append([sourceName,nameValuePair[primaryKey],weightSource,weightFound, relationship, "TBD", risk,userMail,userManagerMail])
    elif(backupKey in nameValuePair.keys()):
        foundList.append(nameValuePair[backupKey])
        edgesList.append([sourceName,nameValuePair[backupKey],weightSource,weightFound, relationship,"TBD", risk,userMail,userManagerMail])
    else:
        defaultList.append(nameValuePair)
        edgesList.append([sourceName,nameValuePair,weightSource,weightNotFound, relationship,"TBD", risk, userMail,userManagerMail])
    
    return (foundList, defaultList, edgesList)

def RiskComparer(current, new):
    '''Compares current risk to new risk, returns the higher'''

    highest = 'Unknown'
    if(MapRiskToValue(current) > MapRiskToValue(new)):
        highest = current
    else:
        highest = new

    return highest

def MapRiskToValue(risk):
    '''Mapper for Risk to a numeric value'''
    dict = {'Unknown': -1, 'None': 0, 'VeryLow': 1, 'Low': 3, 'Medium': 5, 'High': 7, 'VeryHigh': 9, 'Critical': 11}

    return dict[risk] if risk in dict else -1

def filterDataFrameAndCreateList(name_dropdown, file_df, defaultFilter):
    '''Create a filtered list from given edge data CSV file and ipywidgets.Dropdown selection'''

    items = []
    #file_df_filtered = filterHelper(name_dropdown, file_df, defaultFilter)
    for index, row in file_df.iterrows():
        items.append(row)
    return items    

def htmlTreeParser(list):
    '''Given a list of edge data, parse out the items and format for use in HTML Tabulator Tree display'''

    str = ''
    dict = {}
    highestWeight = 0
    highestRisk = 'None'

    # Load the dictionary with all the values
    for item in list:
        key_ls = LongShort(f'{item["Source"]}({item["SourceId"]})',item["Source"])
        target_ls = LongShort(f'{item["Target"]}({item["TargetId"]})',item["Target"])
        dict.setdefault((jsonpickle.encode(key_ls, unpicklable=False).replace('"',"'"),item["SourceType"]), []).append((jsonpickle.encode(target_ls, unpicklable=False).replace('"',"'"), item["TargetType"],item["Relationship"], item["Reason"], item["Risk"], item["TargetWeight"], item["UserMail"], item["UserManagerMail"]))
    
    # Once all values found, trim the dictionary to unique values
    for key,value in dict.items():
        dict[key] = set(value)

    # Format all dictionary values for display with HTML Tabulator tree structure
    for key, value in dict.items():
        highestWeight = 0
        highestRisk = 'None'
        userMail = ''
        userManagerMail = ''
        reason = ''
        childrenString = ''
        relationship = ''
        target_type = ''
        for item in value:
            highestRisk = RiskComparer(highestRisk, item[4])
            highestWeight = highestWeight if (highestWeight > (2*item[5])) else (2*item[5])
            if(not(isNaN(item[1]))):
                target_type = item[1]
            if(not(isNaN(item[2]))):
                relationship = item[2]
            if(not(isNaN(item[3]))):
                reason =  item[3]
            if(not(isNaN(item[6]))):
                userMail =  item[6]
            if(not(isNaN(item[7]))):
                userManagerMail =  item[7]
            childrenString += '{Source:"' + item[0] + '", type:"' + target_type + '", relationship:"' + relationship + '", reason:"' + reason + '", risk:"' + item[4] + '", weight:"' + (f"{2*item[5]}") + '", overall:"' + (f"{MapRiskToValue(item[4]) * 8*item[5]}") + '"},'
        str += '{Source:"' + key[0] + '",type:"' + key[1] + '", risk:"' + highestRisk + '", weight:"' + (f"{highestWeight}") + '", overall:"' + (f"{MapRiskToValue(highestRisk) * highestWeight}") + '", mailLink:"' + userMail + '", managerMailLink:"' + userManagerMail + '", _children:[' + childrenString + ']},'

    return str

def isNaN(string):
    '''Check if a value is NaN'''

    return string != string


def getTreeTemplate():
    '''Template for HTML tree view'''

    treeTemplate = '''<link href="https://unpkg.com/tabulator-tables@5.0.6/dist/css/tabulator.min.css" rel="stylesheet">
<script type="text/javascript" src="https://unpkg.com/tabulator-tables@5.0.6/dist/js/tabulator.min.js"></script>
<link href='https://unpkg.com/tabulator-tables@5.0.6//dist/css/tabulator_midnight.min.css' rel='stylesheet'>
<script src="https://cdnjs.cloudflare.com/ajax/libs/jspdf/2.4.0/jspdf.umd.min.js"></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/jspdf-autotable/3.5.20/jspdf.plugin.autotable.min.js"></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/html2canvas/1.3.2/html2canvas.min.js"></script>
<div>
    <input id='reactivity-includeIds', type='checkbox' name='ids'>
    <label for='ids'> Verbose</label>
</div>
<div>
    <button id="download-csv">Download CSV</button>
    <button id="download-pdf">Download PDF</button>
</div>

<div id="senserva-table"></div>

<script>

    var useLongText = false;

document.getElementById('reactivity-includeIds').addEventListener('click', function () {{
        var checkbox = document.getElementById('reactivity-includeIds');
        useLongText = checkbox.checked;
        table.toggleColumn('type');
        table.replaceData(tableDataNested); 
    }});

document.getElementById("download-csv").addEventListener("click", function(){{
            body = [];
            cols = ['Interactions', 'Relationship', 'Reason', 'Risk', 'Impact', 'Overall', 'Mail Link'];
            if (useLongText) {{
                cols = ['Interactions', 'Object Type', 'Relationship', 'Reason', 'Risk', 'Impact', 'Overall', 'Mail Link'];
            }}
            body = parseRows(table.getData(), body);

            var csvContent = "data:text/csv;charset=utf-8,"

            csvContent += cols.join(",") + "\\r\\n"

            body.forEach(function (rowArray) {{
                var row = rowArray.join(",");
                csvContent += row.replaceAll('#', '') + "\\r\\n"
            }})

            var encodedUri = encodeURI(csvContent);
            var link = document.createElement("a");
            link.setAttribute("href", encodedUri);
            link.setAttribute("download", "SenservaConnectionsContent.csv");
            document.body.appendChild(link);
            link.click();
}});

document.getElementById("download-pdf").addEventListener("click", function(){{
            body = [];
            cols = ['Interactions', 'Relationship', 'Reason', 'Risk', 'Impact', 'Overall', 'Mail Link'];
            if (useLongText) {{
                cols = ['Interactions', 'Object Type', 'Relationship', 'Reason', 'Risk', 'Impact', 'Overall', 'Mail Link'];
            }}
            body = parseRows(table.getData(), body);
            const doc = new jspdf.jsPDF();

            doc.autoTable(cols, body);
            doc.save("SenservaConnectionsContent.pdf");
}});

// non list search using text
    function customHeaderFilterForString(headerValue, rowValue, rowData, filterParams) {{
        //headerValue - the value of the header filter element
        //rowValue - the value of the column in this row
        //rowData - the data for the row being filtered
        //filterParams - params object passed to the headerFilterFuncParams property
        if (headerValue) {{
            var testString = headerValue.toLowerCase();
            if (rowValue) {{
                if (testString.startsWith('!')) {{
                    if(!rowValue.toLowerCase().includes(testString.substring(1))) {{
                        return true;
                    }}
                }}
                else {{
                    if (rowValue.toLowerCase().includes(testString)) {{
                        return true;
                    }}
                }}
            }}

            if (rowData && rowData._children) {{
                var found = false
                rowData._children.forEach(function (element) {{
                    found = found || element[filterParams].toLowerCase().includes(testString)
                }})
                return found
            }}
        }}
        else {{
            return true; // no search
        }}
        return false;
    }}
        
        function checkProperty(row, nameOfProperty) {{
            if (nameOfProperty in row) {{
                return row[nameOfProperty]
            }}
            else {{
                return ' '
            }}
        }}

        function parseRow(row) {{
            var retVal = []
            
            const obj = JSON.parse(checkProperty(row, 'Source').replace(/'/g, '"'));
            if (useLongText) {{
                retVal.push(obj['long'])
                retVal.push(checkProperty(row, 'type'))
            }}
            else {{
                retVal.push(obj['short'])
            }}
            retVal.push(checkProperty(row, 'relationship'))
            retVal.push(checkProperty(row, 'reason'))
            retVal.push(checkProperty(row, 'risk'))
            retVal.push(checkProperty(row, 'weight'))
            retVal.push(checkProperty(row, 'overall'))
            retVal.push(checkProperty(row, 'mailLink'))

            return retVal
        }}

        function parseRows(data, body) {{
            //build table rows
            data.forEach(function (row) {{
                getChildData(row, body);
            }});

            return body;
        }}

        // helper recursive function to fetch child data
        function getChildData(parent, body) {{
            var parent_data = Object.assign({{}}, parent);
            delete parent_data["_children"];
            body.push(parseRow(parent_data));
            if ("_children" in parent) {{
                for (var i = 0, l = parent["_children"].length; i < l; ++i) {{
                    var child = parent["_children"][i];
                    getChildData(child, body);
                }}
            }}
        }}

        

    var populate = function (cell) {{

        try {{
            var inter = cell.getValue().replace(/'/g, '"');
            const obj = JSON.parse(cell.getValue().replace(/'/g, '"'));
                    var s = obj['short'];
                    if (useLongText) {{
                            s = obj['long'];
                    }}
                    var el = cell.getElement();
                    el.style.whiteSpace = 'pre-wrap';
                    el.style.overflow = 'auto';
                    el.style.maxHeight = '100px';
                    return s;
            }}
            catch (err) {{
                    return "<none>";
            }}
    }};

var tableDataNested = [{inputData}];

var table = new Tabulator('#senserva-table', {{
    height: '70%',
    data: tableDataNested,
    dataTree: true,
    dataTreeFilter: false,
    clipboard: true,
    downloadRowRange: 'active',
    pagination:'local',
    paginationSizeSelector:[10, 25, 50, 100],
    paginationSize:10,
    footerElement: "<a href='https:\\www.senserva.com'><button>Powered by Senserva</button></a>",
    columns: [
        {{
            title: 'Interactions',
            field: 'Source',
            width: '20%',
            headerFilter: "input",
            headerFilterFunc: customHeaderFilterForString,
            headerFilterFuncParams: "Source",
            headerTooltip: 'Interactions between AD objects',
            tooltip: 'Interactions',
            formatter:function(cell, formatterParams, onRendered){{
                var data = populate(cell);
                return this.emptyToSpace(this.sanitizeHTML(data));
            }}
        }}, 
        {{
            title: 'Object Type',
            field: 'type',
            width: '10%',
            headerFilter: "input",
            headerFilterFunc: customHeaderFilterForString,
            headerFilterFuncParams: "type",
            headerTooltip: 'The type of the Azure object',
            tooltip: 'Object Type',
            visible: false
        }},
        {{
            title: 'Relationship',
            field: 'relationship',
            width: '15%',
            headerFilter: "input",
            headerFilterFunc: customHeaderFilterForString,
            headerFilterFuncParams: "relationship",
            headerTooltip: 'Relationships between AD objects',
            tooltip: 'Relationships', 
        }},
        {{
            title: 'Insights',
            field: 'reason',
            formatter:function(cell, formatterParams, onRendered){{
                var el = cell.getElement();
                el.style.whiteSpace = 'pre-wrap';
                el.style.overflow = 'auto';
                el.style.maxHeight = '75px';
                return this.emptyToSpace(this.sanitizeHTML(cell.getValue()));
            }},
            width: '20%',
            headerFilter: 'input',
            headerFilterFunc: customHeaderFilterForString,
            headerFilterFuncParams: "reason",
            headerTooltip: 'Insights about AD objects',
            tooltip: 'Insights', 
        }},
        {{
            title: 'Risk',
            field: 'risk',
            width: '10%',
            headerFilter: 'input',
            headerFilterFunc: customHeaderFilterForString,
            headerFilterFuncParams: "risk",
            headerTooltip: 'Senserva evaluated Risk to the network',
            tooltip: 'Risk', 
        }},
        {{
            title: 'Impact',
            field: 'weight',
            align: 'center',
            width: '10%',
            formatter:'progress', 
            formatterParams:{{color:['green', 'orange', 'red']}},
            tooltip:true,
            sorter:"number",
            headerTooltip: 'Senserva evaluated Potential to the network',
            tooltip: 'Impact', 
        }}, 
        {{
            title: 'Overall',
            field: 'overall',
            align: 'center',
            width: '15%',
            formatter:'progress', 
            formatterParams:{{color:['green', 'orange', 'red']}},
            tooltip:true,
            sorter:"number",
            headerTooltip: 'Senserva evaluated Overall threat to the network',
            tooltip: 'Overall', 
        }}, 
        {{
            title: 'Mail Link',
            field: 'mailLink',
            width: '10%',
            headerTooltip: 'Mailto link to AD object and Manager (if available)',
            tooltip: 'Mail Link', 
            formatter:function(cell, formatterParams, onRendered){{
                //cell - the cell component
                //formatterParams - parameters set for the column
                //onRendered - function to call when the formatter has been rendered

                var sendTo = cell.getValue()
                if(sendTo === undefined || sendTo == '')
                {{
                    var el = document.createElement("p")
                    el.innerHTML = "N/A"

                    return el;
                }}

                var ccTo = cell.getData().managerMailLink;
                if(ccTo === undefined)
                {{
                    ccTo = ''
                }}

                var el = document.createElement("a");
                el.setAttribute("href", "mailto://" + cell.getValue() + "?subject=Senserva Connections Data Report Results&cc=" + ccTo)
                el.innerHTML = "Email"
            
                return el; //return the contents of the cell;
            }},
        }},
        {{
            field: 'managerMailLink',
            visible: false
        }}
    ],
    initialSort:[
        {{column:"overall", dir:"desc"}}, 
        {{column:"weight", dir:"desc"}}, 
    ]
}});
</script>'''
    return treeTemplate

def getReportTemplate():
    '''Template for HTML report'''
    reportTemplate = '''<!DOCTYPE html>
<html>
<head>
<div><a href="https:\\www.senserva.com"><img src="https://github.com/Senserva-LLC/SenservaResources/blob/main/HTMLResources/SSL-Logo+Workmark--Vertical--Color.png?raw=true" height="75"></a></div>
<title>Senserva Connections Scanner Report</title>
</head>
<body style="margin-left: auto; margin-right: auto; text-align:center; font-family: sans-serif; padding-right: 10%; padding-left: 10%; ">

<h1>Senserva Connections Scanner Report</h1>
<table>
<tr>
<td>Welcome to the report for your Senserva Connections Scan results! Our team prides itself on clear, accurate reporting of your Connections.</td>
<td>This report will evaluate Users, Groups, Roles, Applications, and Service Principals. Senserva's Pro Scanner will evaluate more metrics such 
as Permissions, PIM Roles, Conditional Access Usage, and more!</td>
<td>The table of results below shows graded connections that exist in your Azure cloud, ordered on risk level. Risk level is calculated based on several factors including as user status, power of role, and type of relationship.</td>
</tr>
</table>
<br/>
<br/>

<div style="maxHeight: 20px; overflow-y: scroll;text-align:left">{table_input}</div>
<br/>
<br/>

<p>Here is a visual network graph of the connections that exist.</p>
<p>Legend<br/> &#9675; - User<br/> &#9674; - Group<br/> &#9651; - Application<br/> &#9734; - Service Principal<br/> &#10133; - Disabled User<br/> &#9633; - Role<br/> &#11041; - PIM Role<br/> &#10683; - Default</p>
<div style="display: flex; justify-content: center;">{graph_input}</div>

</body>
</html>
'''
    return reportTemplate


class LongShort():
    def __init__(self, long, short):
        '''LongShort class, used for objects to have a shorter display value and a longer more informative value, for toggling'''
        self.long = long
        self.short = short


# Load Python libraries that will be used in this notebook to fetch Azure Sentinel data
from azure.loganalytics.models import QueryBody
from azure.mgmt.loganalytics import LogAnalyticsManagementClient
from azure.loganalytics import LogAnalyticsDataClient

# Processing Helpers
import StandaloneApiHelper as SenservaApi

from IPython.display import display, HTML, Markdown
import csv
import pandas as pd
import requests
import webbrowser

# Globals
# We use a file for data set storage so that we only have to grab once
defaultFilter = '(no filter applied)'
filename = 'edge_data_api.csv'
dropdown = [defaultFilter]

import io
import json
import csv
import sys
import pandas as pd
import ipywidgets
import numpy as np
import matplotlib_inline
import matplotlib as map
import matplotlib.image as mpimg
from bokeh.plotting import figure
from bokeh.resources import CDN
from bokeh.embed import file_html

import tkinter as tk
from tkinter import messagebox
from tkinter import *
from urllib.request import urlopen

def main():

    print('''Senserva Python Graph via API  Copyright (C) 2022  Senserva LLC
    This program comes with ABSOLUTELY NO WARRANTY; 
    This is free software, and is provided AS IS; ''')


    authenticated = False
    master = tk.Tk()
    master.title(string="Senserva Connections Scan")

    URL = "https://github.com/Senserva-LLC/SenservaResources/blob/main/HTMLResources/SSL-Symbol-60x60--Green.png?raw=true"
    u = urlopen(URL)
    raw_data = u.read()
    u.close()

    master.tk.call('wm', 'iconphoto', master._w, tk.PhotoImage(data=raw_data))

    frame1 = tk.Frame(master)
    frame2 = tk.Frame(master)
    frame3 = tk.Frame(master)
    
    while not authenticated:
        
        # Need to do the link first, packing after applying grid management is not available
        label = Label(frame2, text="Senserva PowerShell script", fg="blue", cursor="hand2")
        label.pack()
        label.bind("<Button-1>", lambda e: webbrowser.open_new_tab("https://github.com/Senserva-LLC/SenservaForSentinelSetup"))

        tk.Label(frame1, text="Welcome to the Senserva Connections Scan! This standalone scanner will scan your network and build a report for you.").grid(row=0)
        tk.Label(frame1, text="This report builder uses an Azure AD Application ID and Secret to get its dataset.").grid(row=1)
        tk.Label(frame1, text="You can use the script built by our team to create this App or create your own.").grid(row=2)

        tk.Label(frame3, text="Tenant ID").grid(row=2, sticky=tk.W)
        tk.Label(frame3, text="Application ID").grid(row=3, sticky=tk.W)
        tk.Label(frame3, text="Application Secret").grid(row=4, sticky=tk.W)
    
        e1 = tk.Entry(frame3, width=100)
        e2 = tk.Entry(frame3, width=100)
        e3 = tk.Entry(frame3, width=100)
    
        e1.grid(row=2, column=1)
        e2.grid(row=3, column=1)
        e3.grid(row=4, column=1)

        tk.Button(frame3, 
                    text='Continue to the Scan', 
                    command=master.quit).grid(row=6, 
                                            column=0, 
                                            sticky=tk.E, 
                                            pady=4)

        frame1.pack(padx=3, pady=3)
        frame2.pack(padx=3, pady=3)
        frame3.pack(padx=3, pady=3)

        master.mainloop()
   
        try: 
            SenservaApi.authenticate(e1.get(), e2.get(), e3.get()) 
            authenticated = True
            master.quit()
        except tk.TclError as tclErr:
            master.quit()
            authenticated = True
            return
        except Exception as e:
            master.quit()
            tk.messagebox.showerror(title="Error", message="There was a problem getting authentication with the given information. Please check values and try again\n\n" + str(e))
        
    
    print("Getting Service Principal data...")
    spDict = SenservaApi.servicePrincipalsOwners()
    
    print("Getting Application data...")
    appOwnersDict = SenservaApi.applicationsOwners()
    
    print("Getting Role data...")
    rolesMembersDict = SenservaApi.rolesMembers()
    
    print("Getting User data...")
    userMembers = SenservaApi.usersMemberships()
    userManagers = SenservaApi.usersManagers()
    
    print("Getting Group data...")
    groupsMembersDict = SenservaApi.groupsMemberships()
    groupsOwnersDict = SenservaApi.groupsOwners()

    with open(filename, 'w', newline='') as csvfile:
        filewriter = csv.writer(csvfile, delimiter=',',
                        quotechar='|', quoting=csv.QUOTE_MINIMAL)
        
        print("Processing data...")

        # Pull out data from the Graph queries
        SenservaApi.PluckDataFromQueryResults(userMembers, userManagers, groupsOwnersDict, groupsMembersDict, rolesMembersDict, appOwnersDict, spDict, filewriter) 
        
        print("Processing complete!")

        items = []

    with open(filename, newline='') as csvfile:
        file_df = pd.read_csv(csvfile,delimiter=',')

        
        print("Rendering data...")
        
        # Render our data as a Graph
        graphPlot = SenservaApi.RenderGraphData(file_df)
        graphHtml = file_html(graphPlot, CDN, "my plot")

        # Render our data as a Table
        items = SenservaApi.filterDataFrameAndCreateList(dropdown, file_df, defaultFilter)
        inputData = SenservaApi.htmlTreeParser(items)
        tableHtml = SenservaApi.getTreeTemplate().format(inputData = inputData)
        
        print("Rendering complete!")
 
        html = SenservaApi.getReportTemplate().format(table_input=tableHtml, graph_input=graphHtml)

        f = open('SenservaReport.html','w')
        f.write(html)
        f.close()
        webbrowser.open_new_tab('SenservaReport.html')

    print("Complete!")

main()



# Load Python libraries that will be used in this notebook
from azure.common.client_factory import get_client_from_cli_profile
from azure.common.credentials import get_azure_cli_credentials
from azure.loganalytics.models import QueryBody
from azure.mgmt.loganalytics import LogAnalyticsManagementClient
from azure.loganalytics import LogAnalyticsDataClient

# Processing Helpers
import pandas as pd
import json
import ipywidgets
from IPython.display import display, HTML, Markdown
import json
import networkx as nx
import csv
import requests
import jsonpickle
import numpy as np

#Bokeh Visualization Helpers
import bokeh.io
from bokeh.io import output_notebook, show, save
from bokeh.models import Range1d, MultiLine, renderers, Scatter
from bokeh.plotting import figure, from_networkx
from bokeh.palettes import RdYlGn11
from bokeh.transform import linear_cmap
from bokeh.resources import INLINE
bokeh.io.output_notebook(INLINE)

# Functions that will be used in this notebook
def read_config_values(file_path):
    '''This loads pre-generated parameters for Sentinel Workspace. This should be provided to your workspace via a config.json file'''

    with open(file_path) as json_file:
        if json_file:
            json_config = json.load(json_file)
            return (json_config["tenant_id"],
                    json_config["subscription_id"],
                    json_config["resource_group"],
                    json_config["workspace_id"],
                    json_config["workspace_name"],
                    json_config["user_alias"],
                    json_config["user_object_id"])
    return None

def has_valid_token():
    '''Check to see if there is a valid AAD token, otherwise throw exception'''

    try:
        credentials, sub_id = get_azure_cli_credentials()
        creds = credentials._get_cred(resource=None)
        token = creds._token_retriever()[2]
        print("Successfully signed in.")
        return True
    except Exception as ex:
        if "Please run 'az login' to setup account" in str(ex):
            print(str(ex))
            return False
        elif "AADSTS70043: The refresh token has expired" in str(ex):
            message = "**The refresh token has expired. <br> Please continue your login process. Then: <br> 1. If you plan to run multiple notebooks on the same compute instance today, you may restart the compute instance by clicking 'Compute' on left menu, then select the instance, clicking 'Restart'; <br> 2. Otherwise, you may just restart the kernel from top menu. <br> Finally, close and re-load the notebook, then re-run cells one by one from the top.**"
            display(Markdown(message))
            return False
    except:
        print("Please restart the kernel, and run 'az login'.")
        return False

def process_result(result):
    '''This function processes data returned from Azure LogAnalyticsDataClient, it returns pandas DataFrame'''

    json_result = result.as_dict()
    cols = pd.json_normalize(json_result['tables'][0], 'columns')
    final_result = pd.json_normalize(json_result['tables'][0], 'rows')
    if final_result.shape[0] != 0:
        final_result.columns = cols.name

    return final_result
    
def SenservaPermissionQuery(tableName):
    '''Query for finding relationships from the Senserva Scanner results. 
    If you have set up multiple workspaces, provide a Sentinel alias to query both functions'''

    where_clause = "{0} | where TimeGenerated > ago(7d)| where (ControlName_s == 'ServicePrincipalPermissionGrantTenant' or ControlName_s == 'ApplicationPermissionGrantTenant' or ControlName_s == 'ServicePrincipalMembership' or ControlName_s == 'UserMembers' or ControlName_s =='UserOwners' or ControlName_s =='GroupMembers' or ControlName_s == 'UserManagers') | extend values =tostring(parse_json(Value_s)) | extend JSON = todynamic(values) | order by TimeGenerated desc"
    return where_clause.format(tableName)

def SenservaPermissionUebaQuery():
    '''TODO: Docstring'''
    
    #where_clause = " let userId = '{0}';UserAccessAnalytics | where TimeGenerated > ago(7d)| where SourceEntityId == userId and AccessType == 'RBAC' | order by TimeGenerated desc| extend RbacName= TargetEntityName, RbacId = TargetEntityId, RbacObject = TargetEntityType, RbacRole=AccessLevel | distinct RbacName, RbacId, RbacObject, RbacRole"
    
    where_clause = "UserAccessAnalytics | where TimeGenerated > ago(7d)| where AccessType == 'RBAC' | order by TimeGenerated desc| extend SourceId=SourceEntityId,RbacName= TargetEntityName, RbacId = TargetEntityId, RbacObject = TargetEntityType, RbacRole=AccessLevel | distinct SourceId,RbacName, RbacId, RbacObject, RbacRole"

    return where_clause

def PluckDataFromQueryResults(query_result, names_list, csvFileWriter):
    '''Given Senserva query results, pick out the data we want. 
    The data will be written as tuples to a given CSV file writer object. 
    The format of each line of the file will be [Source,Target,relationship,weight].
    All unique objects will be colelcted and returned as a dropdown menu, to allow for user selected filtering'''

    # Set up some values
    name_user = 'User'
    name_disabled_user = 'DisabledUser'
    name_group = 'Group'
    name_role = 'Role'
    name_all_users = 'All Users'
    name_unknown = 'Unknown'

    type_user = 'User'
    type_disabled_user = 'Disabled User'
    type_service_principal = 'Service Principal'
    type_group = 'Group'
    type_application = 'Application'
    type_role = 'Role'
    type_pim_role = 'PIM Role'
    type_unknown = 'Unknown'

    weight_user = '1'
    weight_disabled_user = '8'
    weight_group = '4'
    weight_role = '5'
    weight_application = '6'
    weight_all_users = '3'
    weight_unknown = '10'

    memberRelationship = 'Member'
    indirectMemberRelationship = 'Indirect Member'
    pimMemberRelationship = 'PIM Member'
    ownerRelationship = 'Owner'
    permissionRelationship = 'Permission to Use'
    managerRelationship = 'Manager'

    # Loop through our results and process the relationships
    try:
        csvFileWriter.writerow(['Source', 'SourceId', 'SourceType', 'Target', 'TargetId', 'TargetType', 'SourceWeight', 'TargetWeight', 'Relationship', 'Reason', 'Risk', 'UserMail', 'UserManagerMail'])

        for index,value in query_result.JSON.items():
            JSONItems = json.loads(json.dumps(value)) # This is to handle for if json object comes back with properties wrapped in single quotes
            targetName = "Unknown"
            userList = []
            disabledUserList = []
            unknownObjectList = []
            directGroupList = []
            directRoleList = []
            directApplicationList = []
            directServicePrincipalList = []
            directUnknownList = []
            indirectGroupList = []
            indirectRoleList = []
            indirectUnknownList = []
            edges = []

            userMail = ""
            userManagerMail = ""
            object_id = ""
            reason = ""
            for x in JSONItems:
                if 'Tag' in x and x['Tag'] == "UserMail":
                    userMail = x['Val']
                if 'Tag' in x and x['Tag'] == "UserManagerMail":
                    userManagerMail = x['Val']
                if 'Tag' in x and (x['Tag'] == "UserId" or x['Tag'] == 'ServicePrincipalId'):
                    object_id = x['Val']
                if 'Tag' in x and x['Tag'] == "Reason":
                    reason = x['Val']
                
            for x in JSONItems:
                if 'Tag' in x:

                    # Find our target that we are working with
                    if(x['Tag'] == "ApplicationName" or x['Tag'] =="ServicePrincipalName" or x['Tag'] =="UserName"):
                        targetName = x['Val']
                        names_list.append(targetName)
                    
                    
                    # We have a Service Principal that has been granted consent for the tenant by a tenant admin
                    if(x['Tag'] == "ServicePrincipalTenantPermissionGrantList"):
                        edges.append([targetName,object_id,type_service_principal,name_all_users,"",type_user,weight_application,weight_all_users, permissionRelationship,reason,MapRisk(query_result.Risk_s[index]),userMail,userManagerMail])

                    # We have a Service Principal that has been granted consent for a particular user
                    if(x['Tag'] == "ServicePrincipalUserPermissionGrantList"):
                        for nameValuePair in x['Val']:
                            userList, unknownObjectList, edges = ParseQueryResultHelper(nameValuePair, userList, unknownObjectList, edges, 'Name', 'Value', targetName, object_id, type_service_principal, type_user, weight_application, weight_user, weight_unknown, permissionRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)
                    
                    # We have a Service Principal that has been granted consent for a particular disabled user
                    if(x['Tag'] == "ServicePrincipalDisabledUserPermissionGrantList"):
                        for nameValuePair in x['Val']:
                            disabledUserList, unknownObjectList, edges = ParseQueryResultHelper(nameValuePair, disabledUserList, directUnknownList, edges, 'Name', 'Value', targetName, object_id, type_service_principal, type_disabled_user, weight_application, weight_disabled_user, weight_unknown, permissionRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that has been granted consent for an AAD Object that is not recognized
                    if(x['Tag'] == "ServicePrincipalUnknownObjectPermissionGrantList"):
                        for nameValuePair in x['Val']:
                            unknownObjectList, unknownObjectList, edges = ParseQueryResultHelper(nameValuePair, unknownObjectList, unknownObjectList, edges, 'Name', 'Value', targetName, object_id, type_service_principal, type_unknown, weight_application, weight_unknown, weight_unknown, permissionRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that is directly assigned as a member of an AAD Group
                    if(x['Tag'] == "ServicePrincipalDirectGroupMember"):
                        for nameValuePair in x['Val']:
                            directGroupList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directGroupList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_group, weight_application, weight_group, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that is directly assigned an AAD role membership
                    if(x['Tag'] == "ServicePrincipalDirectRoleMember"):
                        for nameValuePair in x['Val']:
                            directRoleList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directRoleList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_role, weight_application, weight_role, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that is directly assigned membership to an unknown AAD object
                    if(x['Tag'] == "ServicePrincipalDirectUnknownMember"):
                        for nameValuePair in x['Val']:
                            directUnknownList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directUnknownList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_unknown, weight_application, weight_unknown, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that has indirect membership to a group e.g. SP is member of Group A, Group A is a member of Group B
                    if(x['Tag'] == "ServicePrincipalIndirectGroupMember"):
                        for nameValuePair in x['Val']:
                            indirectGroupList, indirectUnknownList, edges = ParseQueryResultHelper(nameValuePair, indirectGroupList, indirectUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_group, weight_application, weight_group, weight_unknown, indirectMemberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that has indirect role membership e.g. SP is member of Group A, Group A is assigned Role B
                    if(x['Tag'] == "ServicePrincipalIndirectRoleMember"):
                        for nameValuePair in x['Val']:
                            indirectRoleList, indirectUnknownList, edges = ParseQueryResultHelper(nameValuePair, indirectRoleList, indirectUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_role, weight_application, weight_role, weight_unknown, indirectMemberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a Service Principal that has indirect membership to an unknown AAD object
                    if(x['Tag'] == "ServicePrincipalIndirectUnknownMember"):
                        for nameValuePair in x['Val']:
                            indirectUnknownList, indirectUnknownList, edges = ParseQueryResultHelper(nameValuePair, indirectUnknownList, indirectUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_unknown, weight_application, weight_unknown, weight_unknown, indirectMemberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a member of an AAD Group
                    if(x['Tag'] == "PimRoleMemberUser"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directRoleList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directGroupList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_pim_role, type_user, weight_user, weight_group, weight_unknown, pimMemberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a member of an AAD Group
                    if(x['Tag'] == "GroupMemberUser"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directGroupList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directGroupList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_group, type_user, weight_user, weight_group, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a assigned a role membership
                    if(x['Tag'] == "RoleMemberUser"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directRoleList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directRoleList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_role, type_user, weight_user, weight_role, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a group that is a assigned a group membership
                    if(x['Tag'] == "GroupMemberGroup"):
                        directGroupList.append(targetName)
                        for nameValuePair in x['Val']:
                            directGroupList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directGroupList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_group, type_group, weight_group, weight_group, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a group that is a assigned a role membership
                    if(x['Tag'] == "GroupMemberRole"):
                        directGroupList.append(targetName)
                        for nameValuePair in x['Val']:
                            directRoleList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directRoleList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_group, type_role, weight_group, weight_role, weight_unknown, memberRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a assigned a group ownership
                    if(x['Tag'] == "GroupOwners"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directGroupList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directGroupList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_group, type_user, weight_group, weight_user, weight_unknown, ownerRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a assigned an application ownership
                    if(x['Tag'] == "ApplicationOwner"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directApplicationList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directApplicationList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_application, type_user, weight_application, weight_user, weight_unknown, ownerRelationship, reason, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that is a assigned a service principal ownership
                    if(x['Tag'] == "ServicePrincipalOwner"):
                        userList.append(targetName)
                        for nameValuePair in x['Val']:
                            directServicePrincipalList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, directServicePrincipalList, directUnknownList, edges, 'Name', 'Id', targetName, object_id, type_service_principal, type_user, weight_application, weight_user, weight_unknown, reason, ownerRelationship, query_result.Risk_s[index],userMail,userManagerMail)

                    # We have a user that has a Manager
                    if(x['Tag'] == "UserManagerName"):
                        userList.append(targetName)
                        edges.append([targetName,object_id,type_user,x['Val'],"",type_user,weight_user,weight_user, managerRelationship,reason,MapRisk(query_result.Risk_s[index]),userMail,userManagerMail])
                        # for nameValuePair in x['Val']:
                        #     userList, directUnknownList, edges = ParseQueryResultHelper(nameValuePair, userList, directUnknownList, edges, 'Name', 'Id', targetName, weight_user, weight_user, weight_unknown, managerRelationship, query_result.Risk_s[index],userMail,userManagerMail)

                else:
                    continue
            for user in userList:   
                names_list.append(str(user))
            for disabledUser in disabledUserList:
                names_list.append(str(disabledUser))
            for unknownObject in unknownObjectList:
                names_list.append(str(unknownObject))
            for directGroup in directGroupList:
                names_list.append(str(directGroup))
            for directRole in directRoleList:
                names_list.append(str(directRole))
            for directApplication in directApplicationList:
                names_list.append(str(directApplication))
            for directServicePrincipal in directServicePrincipalList:
                names_list.append(str(directServicePrincipal))
            for directUnknown in directUnknownList:
                names_list.append(str(directUnknown))
            for indirectGroup in indirectGroupList:
                names_list.append(str(indirectGroup))
            for indirectRole in indirectRoleList:
                names_list.append(str(indirectRole))
            for indirectUnknown in indirectUnknownList:
                names_list.append(str(indirectUnknown))

            # Write our edges to the file for later use
            for edge in edges:
                csvFileWriter.writerow(edge)
                
                
        # Take our list of objects and make a dropdown list to use for filtering
        names = sorted(set(names_list))
        name_dropdown = ipywidgets.Dropdown(options=names, description='Objects:')
        display(name_dropdown)

        return name_dropdown

    except AttributeError:
        print('No data found from the query')

    return


def ParseQueryResultHelper(nameValuePair, foundList, defaultList, edgesList, primaryKey, backupKey, sourceName, source_id, source_type, target_type,weightSource, weightFound, weightNotFound, relationship, reason, risk, userMail,userManagerMail):
    '''Helper to parse through query results'''

    if(primaryKey in nameValuePair.keys()):
        foundList.append(nameValuePair[primaryKey])
        edgesList.append([sourceName,source_id,source_type,nameValuePair[primaryKey],"",target_type,weightSource,weightFound, relationship, reason, MapRisk(risk),userMail,userManagerMail])
    elif(backupKey in nameValuePair.keys()):
        foundList.append(nameValuePair[backupKey])
        edgesList.append([sourceName,source_id,source_type,nameValuePair[backupKey],"",target_type,weightSource,weightFound, relationship, reason, MapRisk(risk),userMail,userManagerMail])
    else:
        defaultList.append(nameValuePair)
        edgesList.append([sourceName,source_id,source_type,nameValuePair,"",target_type,weightSource,weightNotFound, relationship, reason, MapRisk(risk),userMail,userManagerMail])
    
    return (foundList, defaultList, edgesList)

def MapRisk(risk):
    '''Map the Risk Value to a Name'''

    result = ''
    if(risk == 0):
        result = 'None'
    elif(risk == 1):
        result = 'Very Low'
    elif(risk == 5):
        result = 'Low'
    elif(risk == 10):
        result = 'Medium'
    elif(risk == 20):
        result = 'High'
    elif(risk == 30):
        result = 'Very High'
    elif(risk == 50):
        result = 'Critical'
    else:
        result = 'Unknown'

    return result

def glyph_map_helper(item:str):
    '''Mapper for glyphs to type in Graph Renderer'''
    dict = {'User': "circle", 'Group': "diamond", 'Application': "triangle", 'Service Principal': "star", 'Disabled User': "plus", 'PIM Role': "hex", 'Role': "square", 'Default': "circle_x"}

    return dict[item] if item in dict else dict["Default"]

def RenderGraphData(file_df):
    '''Given a Pandas data frame from a CSV file, parse the data out, process it for rendering with a Networkx Graph.
    Each line in Data frame is expected to be in form [Source,Target,SourceWeight,TargetWeight,Relationship,Risk]'''

    # Set up the Graph 
    G = nx.Graph
    keys = {}
    values = {}
    modularity_class = {}
    modularity_color = {}
    modularity_glyph = {}
    relationship = {}
    risk = {}
    target_id = {}
    target_type = {}

    # Process the passed in data frame and make nodes/edges data for graph
    # Use a data frame because networkx expects a data frame
    G = nx.from_pandas_edgelist(file_df, 'Source','Target', True)
    for index, row in file_df.iterrows():
        # SSame node might appear multiple times, find and use only highest value
        if(row['Source'] in keys):
            keys[row['Source']] = max(int(row['SourceWeight']), int(keys[row['Source']]))
        else:
            keys[row['Source']] = int(row['SourceWeight'])

        if(row['Target'] in keys):
            keys[row['Target']] = max(int(row['TargetWeight']), int(keys[row['Target']]))
        else:
            keys[row['Target']] = int(row['TargetWeight'])

        modularity_class[row['Source']] = keys[row['Source']]
        modularity_color[row['Source']] = RdYlGn11[keys[row['Target']]]
        modularity_class[row['Target']] = keys[row['Source']]
        modularity_color[row['Target']] = RdYlGn11[keys[row['Target']]]
        modularity_glyph[row['Source']] = glyph_map_helper(row['SourceType'])
        modularity_glyph[row['Target']] = glyph_map_helper(row['TargetType'])
        relationship[row['Target']] = row['Relationship']
        target_id[row['Source']] = row['SourceId']
        target_id[row['Target']] = row['TargetId']
        target_type[row['Source']] = row['SourceType']
        target_type[row['Target']] = row['TargetType']

        if(row['Source'] in risk.keys()):
            risk[row['Source']] = RiskComparer(risk[row['Source']], row['Risk'])
        else:
            risk[row['Source']] = row['Risk']

        if(row['Target'] in risk.keys()):
            risk[row['Target']] = RiskComparer(risk[row['Target']], row['Risk'])
        else:
            risk[row['Target']] = row['Risk']

    degrees = dict(nx.degree(G))
    nx.set_node_attributes(G, degrees, 'degree')

    # Set the node size after we have processed the data
    number_to_adjust_by = 10
    adjusted_node_size = dict([(node, degree+number_to_adjust_by) for node, degree in nx.degree(G)])
    nx.set_node_attributes(G, adjusted_node_size, 'adjusted_node_size')

    # Add modularity class, color, and relationship as attributes from the network above
    nx.set_node_attributes(G, modularity_class, 'modularity_class')
    nx.set_node_attributes(G, modularity_color, 'modularity_color')
    nx.set_node_attributes(G, relationship, 'relationship')
    nx.set_node_attributes(G, risk, 'risk')
    nx.set_node_attributes(G, target_id, 'target_id')
    nx.set_node_attributes(G, target_type, 'target_type')
    nx.set_node_attributes(G, modularity_glyph, 'marker_type')



    #Choose attributes from G network to size and color by — setting manual size (e.g. 10) or color (e.g. 'skyblue') also allowed
    size_by_this_attribute = 'adjusted_node_size'
    color_by_this_attribute = 'modularity_color'
    marker_by_this_attribute = 'marker_type'

    #Choose a title!
    title = 'Needle in the Haystack'

    #Establish which categories will appear when hovering over each node
    HOVER_TOOLTIPS = [
        ("Azure Name", "@index"),
        ("Azure ID", "@target_id"),
        ("Azure Object Type", "@target_type"),
        ("Number of Connections", "@degree"),
        ("Risk", "@risk"),
    ]

    #Create a plot — set dimensions, toolbar, and title
    plot = figure(tooltips = HOVER_TOOLTIPS,
                tools="pan,wheel_zoom,save,reset, tap", active_scroll='wheel_zoom',
                x_range=Range1d(-10.1, 10.1), y_range=Range1d(-10.1, 10.1), title=title)

    # Create a network graph object
    # https://networkx.github.io/documentation/networkx-1.9/reference/generated/networkx.drawing.layout.spring_layout.html
    network_graph = from_networkx(G, nx.spring_layout, scale=10, center=(0, 0))

    #Set node sizes and colors according to node degree (color as category from attribute)
    network_graph.node_renderer.glyph = Scatter(size=size_by_this_attribute, fill_color=color_by_this_attribute, marker=marker_by_this_attribute)
    
    # Set edge opacity and width
    network_graph.edge_renderer.glyph = MultiLine(line_alpha=0.5, line_width=1)

    # Render the graph
    plot.renderers.append(network_graph)
    
    return plot



def SenservaLocationsQuery(tableName):
    '''Query for finding locations from the Senserva Scanner results. 
    If you have set up multiple workspaces, provide a Sentinel alias to query both functions'''

    where_clause = "{0} | where TimeGenerated > ago(7d)| where ControlName_s == 'Locations'| order by TimeGenerated desc| extend values = tostring(parse_json(Value_s)) | extend JSON = todynamic(values)"
    return where_clause.format(tableName)


def PluckDataFromLocationQueryResults(query_result): #, names_list, csvFileWriter):
    '''TODO: Fill in the docstring'''

    geoList = []
    ids = []
    names = []
    risks = []
    permanentRoles = []
    pimRoles = []
    signInStatus = []
    deviceInformation = []

    try:
        for index,value in query_result.JSON.items():
            JSONItems = json.loads(value)
            current_geopoint = ""
            current_id = ""
            current_name = ""
            current_perm_role = ""
            current_pim_role = ""
            current_sign_in_status = ""
            current_device_info = ""

            for x in JSONItems:
                if 'Tag' in x:
                    if(x['Tag'] == "UserName"):
                        current_name = x['Val']

                    if(x['Tag'] == "UserId"):
                        current_id = x['Val']
                    
                    if(x['Tag'] == "Role"):
                        current_perm_role = x['Val']

                    if(x['Tag'] == "PimRole"):
                        current_pim_role = x['Val']
                    
                    if(x['Tag'] == "SignInStatus"):
                        current_sign_in_status = x['Val']

                    if(x['Tag'] == "UserDeviceDetail"):
                        current_device_info = x['Val']
                        
                    if(x['Tag'] == "LocationCoordinatesHash"):
                        current_geopoint = pygeohash.decode_exactly(x['Val'])
            
            geoList.append(current_geopoint)
            ids.append(current_id)
            names.append(current_name)
            risks.append(query_result.Risk_s[index])
            permanentRoles.append(current_perm_role)
            pimRoles.append(current_pim_role)
            signInStatus.append(current_sign_in_status)
            deviceInformation.append(current_device_info)
                

                    

    except AttributeError:
        print('No data found from the query')   
        
            
    return (geoList,ids,names,risks,permanentRoles,pimRoles,signInStatus,deviceInformation)

def convertDataFrameToList(file_df):
    '''Create a filtered list from given edge data CSV file and ipywidgets.Dropdown selection'''

    items = []
    for index, row in file_df.iterrows():
        items.append(row)
    return items

def htmlTableParser(list):
    '''Given a list of edge data, parse out the items and format for use in HTML Tabulator Table display'''

    str = ''

    for item in list:
        str += '{Source:"' + item["Source"] + '", Target:"' + item["Target"] + '", relationship:"' + item["Relationship"] + '", risk:"' + item["Risk"] + '", weight:"' + f"{item['TargetWeight']}" + '"},'
    
    return str

def htmlTableParser(list):
    '''Given a list of edge data, parse out the items and format for use in HTML Tabulator Table display'''

    str = ''
    dict = {}

    # Load the dictionary with all the values
    for item in list:
        dict.setdefault(item["Source"], []).append((item["Target"], item["Relationship"], item["Risk"], item["TargetWeight"]))

    # Once all values found, trim the dictionary to unique values
    for key,value in dict.items():
        dict[key] = set(value)
    
    # Format all dictionary values for display with HTML Tabulator table structure
    for key, value in dict.items():
        for item in value:
            str += '{Source:"' + key + '", ' + 'Target:"' + item[0] + '", relationship:"' + item[1] + '", risk:"' + item[2] + '", weight:"' + f"{item[3]}" + '"},'

    return str


def htmlTreeParser(list):
    '''Given a list of edge data, parse out the items and format for use in HTML Tabulator Tree display'''

    str = ''
    dict = {}
    highestWeight = 0
    highestRisk = 'None'

    # Load the dictionary with all the values
    for item in list:
        key_ls = LongShort(f'{item["Source"]}({item["SourceId"]})',item["Source"])
        target_ls = LongShort(f'{item["Target"]}({item["TargetId"]})',item["Target"])
        dict.setdefault((jsonpickle.encode(key_ls, unpicklable=False).replace('"',"'"),item["SourceType"]), []).append((jsonpickle.encode(target_ls, unpicklable=False).replace('"',"'"), item["TargetType"],item["Relationship"], item["Reason"], item["Risk"], item["TargetWeight"], item["UserMail"], item["UserManagerMail"]))
    
    # Once all values found, trim the dictionary to unique values
    for key,value in dict.items():
        dict[key] = set(value)

    # Format all dictionary values for display with HTML Tabulator tree structure
    for key, value in dict.items():
        highestWeight = 0
        highestRisk = 'None'
        userMail = ''
        userManagerMail = ''
        reason = ''
        childrenString = ''
        relationship = ''
        target_type = ''
        for item in value:
            highestRisk = RiskComparer(highestRisk, item[4])
            highestWeight = highestWeight if (highestWeight > (2*item[5])) else (2*item[5])
            if(not(isNaN(item[1]))):
                target_type = item[1]
            if(not(isNaN(item[2]))):
                relationship = item[2]
            if(not(isNaN(item[3]))):
                reason =  item[3]
            if(not(isNaN(item[6]))):
                userMail =  item[6]
            if(not(isNaN(item[7]))):
                userManagerMail =  item[7]
            childrenString += '{Source:"' + item[0] + '", type:"' + target_type + '", relationship:"' + relationship + '", reason:"' + reason + '", risk:"' + item[4] + '", weight:"' + (f"{2*item[5]}") + '", overall:"' + (f"{MapRiskToValue(item[4]) * 8*item[5]}") + '"},'
        str += '{Source:"' + key[0] + '",type:"' + key[1] + '", risk:"' + highestRisk + '", weight:"' + (f"{highestWeight}") + '", overall:"' + (f"{MapRiskToValue(highestRisk) * highestWeight}") + '", mailLink:"' + userMail + '", managerMailLink:"' + userManagerMail + '", _children:[' + childrenString + ']},'

    return str

def isNaN(string):
    '''Check whether a given vlaue is NaN'''
    return string != string

def filterHelper(name_dropdown, file_df, defaultFilter):
    '''Based on given edge data CSV file and ipywidgets.Dropdown selection, filter the edge data'''

    if(name_dropdown.value != defaultFilter):
        file_df_filtered = file_df[(file_df.Target == name_dropdown.value)]
        file_df_filtered = file_df_filtered.append(file_df[(file_df.Source == name_dropdown.value)])
        return file_df_filtered
    else:
        return file_df

def filterDataFrameAndCreateList(name_dropdown, file_df, defaultFilter):
    '''Create a filtered list from given edge data CSV file and ipywidgets.Dropdown selection'''

    items = []
    file_df_filtered = filterHelper(name_dropdown, file_df, defaultFilter)
    for index, row in file_df_filtered.iterrows():
        items.append(row)
    return items


def RiskComparer(current, new):
    '''Compares current risk to new risk, returns the higher'''

    highest = 'Unknown'
    if(MapRiskToValue(current) > MapRiskToValue(new)):
        highest = current
    else:
        highest = new

    return highest

def MapRiskToValue(risk):
    '''Mapper for Risk to a numeric value'''
    dict = {'Unknown': -1, 'None': 0, 'VeryLow': 1, 'Low': 2, 'Medium': 3, 'High': 4, 'VeryHigh': 5, 'Critical': 6}

    return dict[risk] if risk in dict else -1

class LongShort():
    def __init__(self, long, short):
        '''LongShort class, used for objects to have a shorter display value and a longer more informative value, for toggling'''
        self.long = long
        self.short = short